package org.baxter.disco.ocr;

//Standard imports
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.lang.invoke.MethodHandles;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.commons.configuration2.INIConfiguration;
import org.apache.commons.configuration2.builder.fluent.Parameters;
import org.apache.commons.configuration2.builder.FileBasedConfigurationBuilder;

/**
 * Facade for saving data out to a file.
 *
 * @author Blizzard Finnegan
 * @version 7.0.0, 10 Mar. 2023
 */
public class DataSaving
{
    /**
     * Path to output file
     */
    private static String outputPath = ConfigFacade.getOutputSaveLocation();

    /**
     * Config API for the output file.
     * This allows the output file to be easily read by both computers
     * and people.
     */
    private static FileBasedConfigurationBuilder<INIConfiguration> OUTPUT_FILE =
        new FileBasedConfigurationBuilder<>(INIConfiguration.class)
        .configure(new Parameters().fileBased().setFile(new File(outputPath)));

    /**
     * Config object for interacting with output values.
     */
    private static INIConfiguration outputValues;

    /**
     * Create local logger
     */
    private static final Logger logger = LoggerFactory.getLogger(MethodHandles.lookup().lookupClass());

    private static int iterationCount;
    private static int passIterationCount;
    private static double sum;
    private static double stdDevIntermediate;
    private static double stdDev;
    private static double median;
    private static final char SEPARATOR = '.';

    static
    {
        try { outputValues = OUTPUT_FILE.getConfiguration(); }
        catch(Exception e){logger.error("Error caught!\nError: ",e); }
    }

    /** 
     * Writes line to XLSX file.
     *
     * @param cycle         What test cycle is being saved to the file 
     * @param serials      Map[String,String] list of camera serials
     *
     * @return Returns whether values were saved successfully.
     */
    public static boolean writeValues(State state)
    {
        boolean output = false;
        if(state.getDataMap().size() > 1)
        {
            state.getDataMap().forEach( (serial,values) -> {
                List<Double> valueList = new ArrayList<>(values.keySet());
                Collections.sort(valueList);
                List<Integer> indexList = new ArrayList<>();
                iterationCount = 0;
                passIterationCount = 0;
                sum = 0;
                stdDevIntermediate = 0;
                median = 0;
                values.forEach( (value,count) -> {
                    iterationCount+= count;
                    if(value > ConfigFacade.getLowerBound() && 
                        value < ConfigFacade.getUpperBound())
                        passIterationCount+=count;
                    sum += value * count;
                });
                int counter = 0;
                for(double value : valueList)
                {
                    counter += values.get(value);
                    indexList.add(counter);
                }
                final double mean = (double)sum/(double)iterationCount;
                int medianIndex = (iterationCount + 1) / 2;
                if(indexList.size() <= 1){ median = valueList.get(0); }
                else
                {
                    for(int i = 0; i < indexList.size(); i++)
                    {
                        int indexGuess = indexList.get(i);
                        if(indexGuess >= medianIndex)
                        {
                            if(iterationCount % 2 == 0 || indexGuess > medianIndex + 1)
                            {
                                logger.trace("Index - 1: " + (i-1));
                                logger.trace("Value passed into array get: " + (i-1 < 0 ? 0 : i-1));
                                median = valueList.get((i-1 < 0 ? 0 : i-1));
                            }
                            else
                                median = (valueList.get((i-1 < 0 ? 0 : i-1)) + valueList.get(i))/2.0;
                        }
                    }
                }

                values.forEach( (value,count) -> {
                    for(int i = 0; i < count; i++)
                        stdDevIntermediate += Math.pow(value-mean,2);
                });

                stdDev = Math.sqrt(stdDevIntermediate / iterationCount);

                String stdDevProperty = serial + SEPARATOR + "standard_deviation";
                String totalProperty = serial + SEPARATOR + "total_iterations";
                String passProperty = serial + SEPARATOR + "pass_iterations";
                String medianProperty = serial + SEPARATOR + "median";

                outputValues.setProperty(totalProperty,iterationCount);
                outputValues.setProperty(passProperty,passIterationCount);
                outputValues.setProperty(stdDevProperty,stdDev);
                outputValues.setProperty(medianProperty,median);
            });
        };
        try
        {
            OUTPUT_FILE.save();
            output = true;
        }
        catch(Exception e){ logger.error("Error caught!\nError: ",e);}
        return output;
    }
}
